from fepydas.datatypes.Data import DataType
WL_nm = DataType("Wavelength","nm")
RS_cm = DataType("Raman Shift","cm-1")
COUNTS = DataType("Counts","#")
TIME_s = DataType("Time","s")
TIME_ns = DataType("Time","ns")
TIME_ps = DataType("Time","ps")
E_eV = DataType("Energy","eV")
ANGLE_au = DataType("Angle","a.u.")
NUMBER = DataType("Number","#")
I_mA = DataType("Current","mA")
P_mW = DataType("Power","mW")
POS_um = DataType("Position","µm")
TEMP_K = DataType("Temperature","K")
